import numpy as np
import pandas as pd


def array_equal(l1, l2):
    if not len(l1) == len(l2):
        return False

    for i in range(len(l1)):
        if np.isnan(l1[i]) and np.isnan(l2[i]):
            continue
        if not l1[i] == l2[i]:
            return False

    return True


def generate_ohlc_df():
    dates = pd.date_range('2018-1-1', '2018-1-31')
    data = [{'open': 1, 'high': 2, 'low': 1, 'close': 2, 'volume': 100}] * len(dates)
    return pd.DataFrame(data=data, index=dates)


def assert_ohlc_df(df):
    assert 'open' in df
    assert 'high' in df
    assert 'low' in df
    assert 'close' in df
    assert 'volume' in df




from .ohlc_feature_producer import FeatureProducerOHLC


class TimeLagsFeatureProducer(FeatureProducerOHLC):
    """
    Produce time lags feature.
    Time lag of feature X traces the previous value of X several time points ago.

    e.g.
    If a stock has close prices of [1, 2, 3, 4, 5, 6, 7], then the 3 time lag of it will be:
    [NaN, NaN, NaN, 1, 2, 3, 4]
    """
    def __init__(self, lags=5, feature='close', feature_label='tlag'):
        """
        Create a new TimeLagsFeatureProducer.

        :param lags: How many previous times lags you want to produce.
        :param feature: On which feature to produce time lags
        """
        FeatureProducerOHLC.__init__(self, feature_label)
        self.lags = lags
        self.feature = feature

    def produce(self, df):
        result = df.copy()

        for i in range(self.lags):
            result['%s_%s_%d' % (self.feature, self.feature_label, i + 1)] = result[self.feature].shift(1 + i)

        return result




from .indices_feature_producer import FeatureProducerIndices


class NasdaqIndexFeatureProducer(FeatureProducerIndices):
    def __init__(self, symbol='QQQ', feature='close', feature_label='nasdaq'):
        FeatureProducerIndices.__init__(self, symbol, feature, feature_label)




import pandas as pd
from feature_util import FeatureProducerPipeline
from feature_util.ohlc import (SimpleMovingAverageFeatureProducer, PercentChangeFeatureProducer, TimeLagsFeatureProducer)


def test_pipeline():
    ohlc_df = pd.DataFrame([
        {'open': 1, 'high': 9, 'low': 1, 'close': 1, 'volume': 100},
        {'open': 1, 'high': 8, 'low': 1, 'close': 2, 'volume': 100},
        {'open': 1, 'high': 7, 'low': 1, 'close': 3, 'volume': 100},
        {'open': 1, 'high': 6, 'low': 1, 'close': 4, 'volume': 100},
        {'open': 1, 'high': 5, 'low': 1, 'close': 5, 'volume': 100},
        {'open': 1, 'high': 4, 'low': 1, 'close': 6, 'volume': 100},
        {'open': 1, 'high': 3, 'low': 1, 'close': 7, 'volume': 100}
    ])

    close_5_sma_producer = SimpleMovingAverageFeatureProducer()
    open_pct_change_producer = PercentChangeFeatureProducer(feature='open')
    volume_1_tlag_producer = TimeLagsFeatureProducer(feature='volume', lags=1)

    pipeline = FeatureProducerPipeline([close_5_sma_producer, open_pct_change_producer, volume_1_tlag_producer])
    result = pipeline.produce(ohlc_df)

    assert 'close_sma_5' in result
    assert 'open_pct_change' in result
    assert 'volume_tlag_1' in result




from ..feature_producer import FeatureProducer


class FeatureProducerOHLC(FeatureProducer):
    """
    This is the base class for OHLC feature producer.
    Given a dataframe with OHLC data, a producer instance will produce
    a new feature based on those data.
    """



from .ohlc_feature_producer import FeatureProducerOHLC
from .moving_average import SimpleMovingAverageFeatureProducer


class BollingerBandsFeatureProducer(FeatureProducerOHLC):
    """
    Produce BB (Bollinger Bands) feature.
    """

    def __init__(self, period=20, feature='close', feature_label='bb'):
        FeatureProducerOHLC.__init__(self, feature_label)
        self.period = period
        self.feature = feature

    def produce(self, df):
        result = df.copy()

        sma = result[self.feature].rolling(self.period).mean()
        std = result[self.feature].rolling(self.period).std()
        assert len(sma) == len(std)

        middle = sma
        upper = middle + (2.0 * std)
        lower = middle - (2.0 * std)

        result['%s_%d_upper' % (self.feature_label, self.period)] = upper
        result['%s_%d_middle' % (self.feature_label, self.period)] = middle
        result['%s_%d_lower' % (self.feature_label, self.period)] = lower

        return result



from .ohlc_feature_producer import FeatureProducerOHLC

from .time_lags import TimeLagsFeatureProducer
from .pct_change import PercentChangeFeatureProducer
from .moving_average import SimpleMovingAverageFeatureProducer
from .moving_average import ExponentialMovingAverageFeatureProducer
from .typical_price import TypicalPriceFeatureProducer
from .cci import CCIFeatureProducer
from .bb import BollingerBandsFeatureProducer
from .force_index import ForceIndexFeatureProducer

from .y_direction import DirectionFeatureProducer



from .ohlc_feature_producer import FeatureProducerOHLC


class ForceIndexFeatureProducer(FeatureProducerOHLC):
    """
    This produces Force Index feature.
    """

    def __init__(self, period=13, feature='close', adjust=True, feature_label='fi'):
        FeatureProducerOHLC.__init__(self, feature_label)
        self.period = period
        self.feature = feature
        self.adjust = adjust

    def produce(self, df):
        result = df.copy()

        fi_1 = result[self.feature].diff() * result['volume']
        fi_period = fi_1.ewm(span=self.period, adjust=self.adjust).mean()

        result['%s_1' % self.feature_label] = fi_1
        result['%s_%d' % (self.feature_label, self.period)] = fi_period

        return result



from .indices_feature_producer import FeatureProducerIndices


class DowJIndexFeatureProducer(FeatureProducerIndices):
    def __init__(self, symbol='DIA', feature='close', feature_label='dowj'):
        FeatureProducerIndices.__init__(self, symbol, feature, feature_label)




from .ohlc_feature_producer import FeatureProducerOHLC


class SimpleMovingAverageFeatureProducer(FeatureProducerOHLC):
    """
    Produce simple moving average (SMA) feature.

    e.g.
    For series [1, 2, 3, 4, 5, 6, 7], a window 2 SMA produces [NaN, 1.5, 2.5, 3.5, 4.5, 5.5, 6.5]
    """

    def __init__(self, period=5, feature='close', min_period=None, feature_label='sma'):
        FeatureProducerOHLC.__init__(self, feature_label)
        self.feature = feature
        self.period = period
        self.min_period = min_period

    def produce(self, df):
        result = df.copy()
        result['%s_%s_%d' % (self.feature, self.feature_label, self.period)] = \
            result[self.feature].rolling(
                window=self.period, min_periods=self.min_period).mean()
        return result


class ExponentialMovingAverageFeatureProducer(FeatureProducerOHLC):
    """
    Produce exponential moving average (EMA) feature.
    """

    def __init__(self, period=5, feature='close', min_period=0, adjust=True, feature_label='ema'):
        FeatureProducerOHLC.__init__(self, feature_label)
        self.feature = feature
        self.period = period
        self.min_period = min_period
        self.adjust = adjust

    def produce(self, df):
        result = df.copy()
        result['%s_%s_%d' % (self.feature, self.feature_label, self.period)] = \
            result[self.feature].ewm(
                span=self.period, min_periods=self.min_period, adjust=self.adjust).mean()
        return result



from .ohlc_feature_producer import FeatureProducerOHLC


class TypicalPriceFeatureProducer(FeatureProducerOHLC):
    """
    Produce typical price feature.
    tp = (High + Low + Close) / 3
    """
    def __init__(self, feature_label='tp'):
        FeatureProducerOHLC.__init__(self, feature_label)

    def produce(self, df):
        result = df.copy()
        high = result['high']
        low = result['low']
        close = result['close']
        result[self.feature_label] = (high + low + close) / 3.0

        return result



from ..feature_producer import FeatureProducer
from datasource.price import get_historical


class FeatureProducerIndices(FeatureProducer):
    """
    This is the base class for indices feature producer.
    It can provide popular indices data.
    """
    def __init__(self, symbol, feature, feature_label):
        """
        :param str feature: Which feature of index to produce (e.g. close)
        :param str feature_label: The label you like for new feature.
        :param str symbol: symbol code for index
        """
        FeatureProducer.__init__(self, feature_label)
        self.feature = feature
        self.symbol = symbol

    def produce(self, df):
        result = df.copy()

        df_index = [str(s)[:10] for s in sorted(df.index.values)]   # use index of given df to determine period
        freq = 'daily'  # TODO determine freq based on index of given df
        sp500_data = get_historical(self.symbol, df_index[0], df_index[-1], freq)[self.feature]
        result[self.feature_label] = sp500_data

        return result



from .util import array_equal
from .util import generate_ohlc_df
from .util import assert_ohlc_df
import numpy as np


def test_array_equal():
    l1 = [1, 2, 3, 4, 5]
    l2 = [1, 2, 3, 4, 5]
    l3 = [1, 2, 3, 4]
    l4 = [np.nan, np.nan, 1, 1]
    l5 = [np.nan, np.nan, 1, 1]

    assert array_equal(l1, l2)
    assert not array_equal(l1, l3)
    assert array_equal(l4, l5)


def test_generate_ohlc_df():
    df = generate_ohlc_df()
    assert_ohlc_df(df)



from .ohlc_feature_producer import FeatureProducerOHLC
from .typical_price import TypicalPriceFeatureProducer


class CCIFeatureProducer(FeatureProducerOHLC):
    """
    Produce the CCI (Commodity channel index) feature.

    Typically it's compared to +/-100, where cci > 100 means overbought and < -100 means oversold.
    """
    def __init__(self, period=20, feature_label='cci', abs_threshold=100):
        """

        :param period: The period for taking sma/mad for typical price
        :param feature_label: The name of feature column
        :param abs_threshold:
        The threshold to produce CCI relative value.
        If not given, the computed CCI valued will be produced directly. Otherwise, only values above abs_threshold or
        below - abs_threshold will be produced in the form of the difference between +/- abs_threshold
        """
        FeatureProducerOHLC.__init__(self, feature_label)
        self.period = period
        self.abs_threshold = abs_threshold  # TODO how we include the +/-100 threshold in the feature?

    def produce(self, df):

        def cci_scale(value):
            threshold = self.abs_threshold

            if abs(threshold) > value > - abs(threshold):
                return 0
            elif value > abs(threshold):
                return value - abs(threshold)
            else:
                return value + abs(threshold)

        result = df.copy()
        # make typical price first
        tp_producer = TypicalPriceFeatureProducer()
        tp = tp_producer.produce(result)['tp']

        tp_sma = tp.rolling(window=self.period).mean()
        tp_abs_sma_diff = (tp - tp_sma).abs()
        tp_mean_abs_devia = tp_abs_sma_diff.rolling(window=self.period).mean()

        cci_value = (tp - tp_sma) / (tp_mean_abs_devia * 0.015)

        if self.abs_threshold is not None:
            cci_value = cci_value.apply(cci_scale)

        result[self.feature_label] = cci_value
        return result



from ..feature_producer import FeatureProducer
from datasource.treasury import get_us_treasury_yield


class USTreasuryYieldFeatureProducer(FeatureProducer):
    """
    This produces US treasury yield data. 
    """

    def __init__(self, maturity='10y', feature_label='us_treasury'):
        """
        Keyword Arguments:
            maturity {str} -- [Maturity period e.g 1m, 10y] (default: {'10y'})
        """
        FeatureProducer.__init__(self, feature_label)
        self.maturity = maturity

    def produce(self, df):
        result = df.copy()

        df_index = [str(s)[:10] for s in sorted(df.index.values)]
        treasury_data = get_us_treasury_yield(
            df_index[0], df_index[-1], maturity=self.maturity)
        result['%s_%s' % (self.feature_label, self.maturity)] = treasury_data

        return result



import numpy as np
from .ohlc_feature_producer import FeatureProducerOHLC


class DirectionFeatureProducer(FeatureProducerOHLC):
    """
    NOTE: This is a y feature. (A feature that should be the training target)

    Produce price moving direction feature. 1 for going up (stay still) and 0 for going down.
    """
    def __init__(self, lags=1, feature='close', feature_label='direction'):
        FeatureProducerOHLC.__init__(self, feature_label)
        self.lags = lags
        self.feature = feature

    def produce(self, df):
        result = df.copy()

        for i in range(self.lags):
            feature_shift = result[self.feature].shift(-(i + 1))
            feature_diff = feature_shift - result[self.feature]
            direction = feature_diff.apply(lambda x: 1 if x >= 0 else 0 if x < 0 else np.nan)

            label = 'Y_%s_%s_%d' % (self.feature, self.feature_label, i + 1)
            result[label] = direction

        return result



from .ohlc_feature_producer import FeatureProducerOHLC

class PercentChangeFeatureProducer(FeatureProducerOHLC):
    """
    Produce percent change feature.

    e.g.
    For series [1, 2, 3, 4, 5], produces [NaN, 1, 0.5, 0.33, 0.25]
    """
    def __init__(self, feature='close', feature_label='pct_change'):
        FeatureProducerOHLC.__init__(self, feature_label)
        self.feature = feature

    def produce(self, df):
        result = df.copy()
        result['%s_%s' % (self.feature, self.feature_label)] = result[self.feature].pct_change()
        return result



class FeatureProducer:
    def __init__(self, feature_label):
        """
        :param feature_label: The label you like for new feature.
        """
        self.feature_label = feature_label

    def produce(self, df):
        """
        Produce a new feature based on given dataframe.

        :param df: A dataframe that is indexed by pd.DatetimeIndex, and has columns of open, high, low, close, volume as float.
        :return: A new dataframe that has the new feature column added.
        """
        pass


class FeatureProducerPipeline(FeatureProducer):
    """
    This gives users the ability to combine several feature producers into a pipeline.
    One can create a list of solid producer instances and put them in the pipeline to construct
    a big new producer.
    """
    def __init__(self, producers):
        FeatureProducer.__init__(self, None)
        self.producers = producers

    def produce(self, df):
        """
        Iterate the given dataframe through all the producers in the list.
        :param df: A dataframe that is indexed by pd.DatetimeIndex, and has columns of open, high, low, close, volume as float.
        :return: A new dataframe that has all the new feature columns produced by each producer
        """
        result = df.copy()
        for p in self.producers:
            result = p.produce(result)

        return result
